/*
 * Copyright (C) 2018 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 * Author: Tomasz Wlostowski <tomasz.wlostowski@cern.ch>
 *
 * SPDX-License-Identifier: LGPL-3.0-or-later
 */


#include <errno.h>
#include <string.h>

#include "mockturtle-rt.h"
#include <mockturtle-framework.h>
#include "wrtd-common.h"
#include "wrtd-fw-common.h"
#include "hw/fd_channel_regs.h"
#include "hw/fd_main_regs.h"

#include "fine-delay-init.h"

#include "wrtd-fw-internals.h"

/**
 * fd_do_reset
 * The reset function (by Tomasz)
 *
 * This function can reset the entire mezzanine (FMC) or just
 * the fine-delay core (CORE).
 * In the reset register 0 means reset, 1 means normal operation.
 */
static void fd_do_reset(int hw_reset)
{
	if (hw_reset) {
		/* clear RSTS_RST_FMC bit, set  RSTS_RST_CORE bit*/
		fd_writel( FD_RSTR_LOCK_W(0xdead) | FD_RSTR_RST_CORE_MASK,
		       FD_REG_RSTR);
		udelay(10000);
		fd_writel( FD_RSTR_LOCK_W(0xdead) | FD_RSTR_RST_CORE_MASK
		       | FD_RSTR_RST_FMC_MASK, FD_REG_RSTR);
		/* TPS3307 supervisor needs time to de-assert master reset */
		mdelay(600);
		return;
	}

	/* clear RSTS_RST_CORE bit, set RSTS_RST_FMC bit */
	fd_writel( FD_RSTR_LOCK_W(0xdead) | FD_RSTR_RST_FMC_MASK,
		  FD_REG_RSTR);
	udelay(1000);
	fd_writel( FD_RSTR_LOCK_W(0xdead) | FD_RSTR_RST_FMC_MASK
	       | FD_RSTR_RST_CORE_MASK, FD_REG_RSTR);
	udelay(1000);
}

/* Some init procedures to be intermixed with subsystems */
int fd_gpio_defaults()
{
	fd_gpio_dir( FD_GPIO_TRIG_INTERNAL, FD_GPIO_OUT);
	fd_gpio_set( FD_GPIO_TRIG_INTERNAL);

	fd_gpio_set( FD_GPIO_OUTPUT_MASK);
	fd_gpio_dir( FD_GPIO_OUTPUT_MASK, FD_GPIO_OUT);

	fd_gpio_dir( FD_GPIO_TERM_EN, FD_GPIO_OUT);
	fd_gpio_clr( FD_GPIO_TERM_EN);
	return 0;
}

int fd_reset_again()
{
	/* Reset the FD core once we have proper reference/TDC clocks */
	fd_do_reset( 0 /* not hw */);

	mdelay(10);

	if (! ( fd_readl(FD_REG_GCR) & FD_GCR_DDR_LOCKED) )
	{
		pr_error("timeout waiting for GCR lock bit\n");
		return -EIO;
	}

	fd_do_reset( 0 /* not hw */);
	return 0;
}

/* FIXME missing all calibration */
int fd_init(void)
{
	int err, ch;


	pr_debug("Initializing the Fine Delay board...\n");

	fd_do_reset(1);

	err = fd_gpio_init();
	if (err)
		return err;

	err = fd_pll_init();
	if (err)
		return err;

	fd_gpio_defaults();

	err = fd_reset_again();
	if (err)
		return err;

	err = fd_acam_init();
	if (err)
		return err;

	for (ch = 1; ch <= FD_CH_NUMBER; ch++)
		fd_gpio_set( FD_GPIO_OUTPUT_EN(ch));

	// todo: read offsets from the EEPROM. I2C should be working OK.

	return 0;
}
