/*
 * SPI access to fine-delay internals
 *
 * Copyright (C) 2012 CERN (www.cern.ch)
 * Author: Tomasz Wlostowski <tomasz.wlostowski@cern.ch>
 * Author: Alessandro Rubini <rubini@gnudd.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2 as published by the Free Software Foundation or, at your
 * option, any later version.
 */

#include "mockturtle-rt.h"

#include "fine-delay-init.h"

static uint32_t fd_mcp_iodir = 0;
static uint32_t fd_mcp_olat = 0;

#define SPI_RETRIES 100

int gpio_writel(int val, int reg)
{
	int rval = fd_spi_xfer( FD_CS_GPIO, 24,
			   0x4e0000 | (reg << 8) | val, NULL);

	return rval;
}

static int gpio_readl(int reg)
{
	uint32_t ret;
	int err;

	err = fd_spi_xfer(FD_CS_GPIO, 24,
			  0x4f0000 | (reg << 8), &ret);

	if (err < 0)
		return err;
	return ret & 0xff;
}

static int gpio_writel_with_retry( int val, int reg)
{
	int retries = SPI_RETRIES, rv;
	while(retries--)
	{
		gpio_writel( val, reg);
		rv = gpio_readl( reg);
		if(rv >= 0 && (rv == val))
		{
			if(SPI_RETRIES-1-retries > 0)
				pr_debug("gpio_writel_with_retry: succeded after %d retries\n",
				       SPI_RETRIES - 1 - retries);
			return 0;
		}
	}
	return -EIO;
}

void fd_gpio_dir(int mask, int dir)
{
	fd_mcp_iodir &= ~mask;
	if (dir == FD_GPIO_IN)
		fd_mcp_iodir |= mask;

	gpio_writel_with_retry( (fd_mcp_iodir & 0xff), FD_MCP_IODIR);
	gpio_writel_with_retry( (fd_mcp_iodir >> 8), FD_MCP_IODIR+1);
}

void fd_gpio_val(int mask, int values)
{

	fd_mcp_olat &= ~mask;
	fd_mcp_olat |= values;

	gpio_writel_with_retry( (fd_mcp_olat & 0xff), FD_MCP_OLAT);
	gpio_writel_with_retry( (fd_mcp_olat >> 8), FD_MCP_OLAT+1);
}

void fd_gpio_set_clr(int mask, int set)
{
	if (set)
		fd_gpio_val(mask, mask);
	else
		fd_gpio_val(mask, 0);
}

int fd_gpio_init()
{
	int i, val;

	fd_mcp_iodir = 0xffff;
	fd_mcp_olat = 0;

	if (gpio_writel(0x00, FD_MCP_IOCON) < 0)
		goto out;

	/* Try to read and write a register to test the SPI connection */
	for (val = 0xaa; val >= 0; val -= 0x11) {
		if (gpio_writel(val, FD_MCP_IPOL) < 0)
			goto out;
		i = gpio_readl( FD_MCP_IPOL);
		if (i < 0)
			goto out;
		if (i != val) {
			pr_error("GPIO comm error (got 0x%x, expected 0x%x)\n", i, val);
			return -EIO;
		}
	}
	/* last time we wrote 0, ok */
	return 0;
out:
	return -EIO;
}
