/*
 * I2C access (on-board EEPROM)
 *
 * Copyright (C) 2012 CERN (www.cern.ch)
 * Author: Tomasz Wlostowski <tomasz.wlostowski@cern.ch>
 * Author: Alessandro Rubini <rubini@gnudd.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2 as published by the Free Software Foundation or, at your
 * option, any later version.
 */

#include "mockturtle-rt.h"

#include "fine-delay-init.h"
#include "hw/fd_main_regs.h"

static void set_sda(int val)
{
	uint32_t reg;

	reg = fd_readl(FD_REG_I2CR) & ~FD_I2CR_SDA_OUT;
	if (val)
		reg |= FD_I2CR_SDA_OUT;
	fd_writel(reg, FD_REG_I2CR);
	udelay(3);
}

static void set_scl(int val)
{
	uint32_t reg;

	reg = fd_readl(FD_REG_I2CR) & ~FD_I2CR_SCL_OUT;
	if (val)
		reg |= FD_I2CR_SCL_OUT;
	fd_writel(reg, FD_REG_I2CR);
	udelay(3);
}

static int get_sda()
{
	return fd_readl(FD_REG_I2CR) & FD_I2CR_SDA_IN ? 1 : 0;
};

static void mi2c_start()
{
	set_sda(0);
	set_scl(0);
}

static void mi2c_stop()
{
	set_sda(0);
	set_scl(1);
	set_sda(1);
}

int mi2c_put_byte(int data)
{
	int i;
	int ack;

	for (i = 0; i < 8; i++, data<<=1) {
		set_sda(data & 0x80);
		set_scl(1);
		set_scl(0);
	}

	set_sda(1);
	set_scl(1);

	ack = get_sda();

	set_scl(0);
	set_sda(0);

	return ack ? -EIO : 0; /* ack low == success */
}

int mi2c_get_byte(unsigned char *data, int sendack)
{
	int i;
	int indata = 0;

	/* assert: scl is low */
	set_scl(0);
	set_sda(1);
	for (i = 0; i < 8; i++) {
		set_scl(1);
		indata <<= 1;
		if (get_sda())
			indata |= 0x01;
		set_scl(0);
	}

	set_sda((sendack ? 0 : 1));
	set_scl(1);
	set_scl(0);
	set_sda(0);

	*data= indata;
	return 0;
}

void mi2c_init()
{
	set_scl(1);
	set_sda(1);
}

void mi2c_scan()
{
	int i;
	for(i = 0; i < 256; i += 2) {
		mi2c_start();
		if(!mi2c_put_byte(i)) 
			pr_debug("Found i2c device at 0x%x\n",
			         i >> 1);
		mi2c_stop();
	}
}

/* FIXME: this is very inefficient: read several bytes in a row instead */
int fd_eeprom_read(int i2c_addr, uint32_t offset,
		void *buf, size_t size)
{
	int i;
	uint8_t *buf8 = buf;
	unsigned char c;

	for(i = 0; i < size; i++) {
		mi2c_start();
		if(mi2c_put_byte(i2c_addr << 1) < 0) {
			mi2c_stop();
			return -EIO;
		}

		mi2c_put_byte((offset >> 8) & 0xff);
		mi2c_put_byte(offset & 0xff);
		offset++;
		mi2c_stop();
		mi2c_start();
		mi2c_put_byte((i2c_addr << 1) | 1);
		mi2c_get_byte(&c, 0);
		*buf8++ = c;
		mi2c_stop();
	}
	return size;
}
