/*
 * Copyright (C) 2018 CERN (www.cern.ch)
 * Author: Dimitris Lampridis <dimitris.lampridis@cern.ch>
 *
 * SPDX-License-Identifier: LGPL-3.0-or-later
 */

#include "mockturtle-rt.h"

#include "wrtd-rt-tdc.h"

#include "hw/fmctdc-direct.h"
#include "hw/tdc_regs.h"
#include "hw/acam_gpx.h"

static struct {
	uint32_t reg;
	uint32_t value;
} acam_config[NB_ACAM_REGS] = {
	{
		0, AR0_ROsc | AR0_HQSel | AR0_TRiseEn(0) |
		AR0_TRiseEn(1) | AR0_TRiseEn(2) |
		AR0_TRiseEn(3) | AR0_TRiseEn(4) |
		AR0_TRiseEn(5) |
		AR0_TFallEn(1) | AR0_TFallEn(2) |
		AR0_TFallEn(3) | AR0_TFallEn(4) | AR0_TFallEn(5)}, {
		1, 0}, {
		2, AR2_IMode | AR2_Disable(6) | AR2_Disable(7) | AR2_Disable(8)}, {
		3, 0}, {
		4, AR4_StartTimer(15) | AR4_EFlagHiZN}, {
		5, AR5_StartOff1(2000)}, {
		6, AR6_Fill(0xfc)}, {
		7, AR7_RefClkDiv(7) | AR7_HSDiv(234) | AR7_NegPhase | AR7_ResAdj}, {
		11, AR11_HFifoErrU(0) | AR11_HFifoErrU(1) |
		AR11_HFifoErrU(2) | AR11_HFifoErrU(3) |
		AR11_HFifoErrU(4) | AR11_HFifoErrU(5) |
		AR11_HFifoErrU(6) | AR11_HFifoErrU(7)}, {
		12, AR12_StartNU | AR12_HFifoE}, {
		14, 0}
};

static inline int acam_is_pll_locked(void)
{
	uint32_t status;

	dp_writel(TDC_CTRL_READ_ACAM_CFG, BASE_DP_TDC_REGS + TDC_REG_CTRL);
	udelay(100);

	status = dp_readl(BASE_DP_TDC_REGS + TDC_REG_ACAM_READBACK(12));

	return !(status & AR12_NotLocked);
}

int tdc_acam_init(void)
{
	int i;

	pr_debug("%s: initializing ACAM TDC...\n", __func__);

	dp_writel(TDC_CTRL_RESET_ACAM, BASE_DP_TDC_REGS + TDC_REG_CTRL);
	udelay(100);

	for (i = 0; i < NB_ACAM_REGS; i++) {
		dp_writel(acam_config[i].value,
			  BASE_DP_TDC_REGS + TDC_REG_ACAM_CONFIG(acam_config[i].reg));
	}

	/* commit ACAM config regs */
	dp_writel(TDC_CTRL_LOAD_ACAM_CFG, BASE_DP_TDC_REGS + TDC_REG_CTRL);
	udelay(100);

	/* and reset the chip (keeps configuration) */
	dp_writel(TDC_CTRL_RESET_ACAM, BASE_DP_TDC_REGS + TDC_REG_CTRL);
	udelay(100);

	/* wait for the ACAM's PLL to lock (2 seconds) */
	mdelay(2000);
	if (acam_is_pll_locked()) {
		pr_debug("%s: ACAM initialization OK.\n",__func__);
		return 0;
	}

	pr_error("%s: ACAM PLL doesn't lock\n", __func__);
	return -EIO;
}

int tdc_time_init(void)
{
	/* program the VCXO DAC to the default calibration value */
	dp_writel(TDC_VCXO_DEFAULT_TUNE, BASE_DP_TDC_REGS + TDC_REG_DAC_TUNE);
	dp_writel(TDC_CTRL_CONFIG_DAC, BASE_DP_TDC_REGS + TDC_REG_CTRL);

	/* set TAI time to zero */
	dp_writel(0, BASE_DP_TDC_REGS + TDC_REG_START_UTC);
	dp_writel(TDC_CTRL_LOAD_UTC, BASE_DP_TDC_REGS + TDC_REG_CTRL);

	return 0;
}

int tdc_enable_termination(int channel, int enable)
{
	uint32_t ien;

	if (channel < FT_CH_1 || channel > FT_NUM_CHANNELS)
		return -EINVAL;

	ien = dp_readl(BASE_DP_TDC_REGS + TDC_REG_INPUT_ENABLE);

	if (enable)
		ien |= (1 << (channel - 1));
	else
		ien &= ~(1 << (channel - 1));

	dp_writel(ien, BASE_DP_TDC_REGS + TDC_REG_INPUT_ENABLE);

	return 0;
}

int tdc_channels_init(void)
{
	int i;

	for (i = FT_CH_1; i <= FT_NUM_CHANNELS; i++) {
		/* termination is off by default */
		tdc_enable_termination(i, 0);
	}

	return 0;
}

void tdc_enable_acquisition(int enable)
{
	uint32_t ien;

	ien = dp_readl(BASE_DP_TDC_REGS + TDC_REG_INPUT_ENABLE);
	if (enable) {
		/* Enable TDC acquisition */
		dp_writel(ien | TDC_INPUT_ENABLE_CH_ALL | TDC_INPUT_ENABLE_FLAG,
			  BASE_DP_TDC_REGS + TDC_REG_INPUT_ENABLE);
		/* Enable ACAM acquisition */
		dp_writel(TDC_CTRL_EN_ACQ, BASE_DP_TDC_REGS + TDC_REG_CTRL);
	} else {
		/* Disable ACAM acquisition */
		dp_writel(TDC_CTRL_DIS_ACQ, BASE_DP_TDC_REGS + TDC_REG_CTRL);
		/* Disable TDC acquisition */
		dp_writel(ien & ~(TDC_INPUT_ENABLE_CH_ALL | TDC_INPUT_ENABLE_FLAG),
			  BASE_DP_TDC_REGS + TDC_REG_INPUT_ENABLE);
	}
}

int tdc_irq_init(void)
{
	/* IRQ coalescing: 40 timestamps or 40 milliseconds */
	dp_writel(40, BASE_DP_TDC_REGS + TDC_REG_IRQ_THRESHOLD);
	dp_writel(40, BASE_DP_TDC_REGS + TDC_REG_IRQ_TIMEOUT);

	return 0;
}

/* FIXME missing all calibration */
int tdc_init(void)
{
	int err;

	pr_debug("%s: Initializing the TDC...\n\r", __func__);

	/* Initialize the TDC FIFO (channels disabled, default dead time) */
	dp_writel(0x0, BASE_DP_TDC_DIRECT + DR_REG_CHAN_ENABLE);
	dp_writel(DEFAULT_DEAD_TIME, BASE_DP_TDC_DIRECT + DR_REG_DEAD_TIME);

	tdc_enable_acquisition(0);

	err = tdc_acam_init();
	if (err)
		return err;

	err = tdc_time_init();
	if (err)
		return err;

	err = tdc_channels_init();
	if (err)
		return err;

	err = tdc_irq_init();
	if (err)
		return err;

	tdc_enable_acquisition(1);

	/* Enable the TDC FIFO */
	dp_writel(0x1f, BASE_DP_TDC_DIRECT + DR_REG_CHAN_ENABLE);

	pr_debug("%s: TDC initialization complete\n\r", __func__);

	return 0;
}
